import enum

__all__ = ['reCAPTCHAType']


class reCAPTCHAType(enum.IntEnum):

    ###
    # Determine the reCAPTCHA type automatically
    #
    #   NOTE: this is a guess as to what 0 means. it is the default in the
    #         official client, but values of 0 are not mentioned in the API docs.
    #
    AUTO = 0

    ###
    # Regular, visible-on-page-load reCAPTCHA v2
    #
    NORMAL = 1

    ###
    # Initially-invisible-until-form-submission reCAPTCHA v2
    #
    INVISIBLE = 2

    ###
    # Score-based reCAPTCHA v3
    #
    V3 = 3

    def __str__(self):
        # This ensures enum values are serialized as their integer values over
        # the wire in POST bodies, rather than their name (e.g. "reCAPTCHAType.NORMAL")
        return str(self.value)



import logging
from typing import TypeVar

T = TypeVar('T')


###
# Add a TRACE log level, if one does not already exist
#
TRACE = logging.getLevelName('TRACE')
if not isinstance(TRACE, int):
    TRACE = 5
    logging.addLevelName(TRACE, 'TRACE')


def getLogger(name: str) -> 'Logger':
    logger = logging.getLogger(name)
    logger.__class__ = Logger
    return logger


class Logger(logging.Logger):
    """Logger supporting TRACE level, and convenient sub-logger creation
    """
    def subLogger(self: T, name: str) -> T:
        return getLogger(f'{self.name}.{name}')

    sub_logger = subLogger

    def trace(self, msg, *args, **kwargs) -> None:
        """
        Log 'msg % args' with severity 'TRACE'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.trace("Houston, we have a %s", "interesting problem", exc_info=1)
        """
        if self.isEnabledFor(TRACE):
            self._log(TRACE, msg, args, **kwargs)


class ClassLoggingProperty:
    """A class-level property declaration exposing a logger with the class's name
    """
    def __init__(self, parent: Logger):
        self.parent = parent

    def __get__(self, obj, cls) -> Logger:
        try:
            return cls._logger
        except AttributeError:
            cls._logger = self.parent.subLogger(cls.__name__)
            return cls._logger



import httpx

__all__ = [
    'ImageTyperzError',
    'AuthenticationFailed',
    'InvalidDomain',
    'InvalidSiteKey',
    'InvalidCaptchaId',
    'NotDecoded',
    'ImageTimedOut',
    'AutomatedQueries',
    'NotInvisible',
    'LimitExceeded',
    'MaximumAttemptsReached',
    'raise_on_error',
]


class ImageTyperzError(Exception):
    pass


class AuthenticationFailed(ImageTyperzError):
    """Provided username and password and/or access key are invalid"""


class InvalidDomain(ImageTyperzError):
    """Wrong domain. Check your domain"""


class InvalidSiteKey(ImageTyperzError):
    """Wrong site key. Check your site key"""


class InvalidCaptchaId(ImageTyperzError):
    """Invalid/nonexistent captcha job ID"""


class NotDecoded(ImageTyperzError):
    """Captcha is still being completed

    This is not really an error, just notifies the captcha is still in progress.
    When you get this, retry after 5 seconds.
    """


class ImageTimedOut(ImageTyperzError):
    """Captcha wasn't completed by workers in time"""


class AutomatedQueries(ImageTyperzError):
    """Proxy sent is blocked temporarily. Change your proxy"""


class NotInvisible(ImageTyperzError):
    """Captcha is not invisible"""


class LimitExceeded(ImageTyperzError):
    """Server is overloaded"""


ERROR_TEXT_EXC_MAP = {
    'AUTHENTICATION_FAILED': AuthenticationFailed,
    'INVALID_DOMAIN': InvalidDomain,
    'INVALID_SITEKEY': InvalidSiteKey,
    'INVALID_CAPTCHA_ID': InvalidCaptchaId,
    'NOT_DECODED': NotDecoded,
    'IMAGE_TIMED_OUT': ImageTimedOut,
    'AUTOMATED_QUERIES': AutomatedQueries,
    'NOT_INVISIBLE': NotInvisible,
    'LIMIT_EXCEED': LimitExceeded,
}


class MaximumAttemptsReached(RuntimeError):
    """The maximum number of attempts has been made; cannot continue"""


def raise_on_error(response: httpx.Response) -> None:
    """Raise the appropriate exception if the response contains an ERROR: ...
    """
    text = response.text
    if 'ERROR:' in text:
        _, err_type = text.split('ERROR:', maxsplit=1)
        err_type = err_type.strip()
        exc_class = ERROR_TEXT_EXC_MAP.get(err_type, ImageTyperzError)
        raise exc_class(err_type)



from typing import Generator
from urllib.parse import parse_qsl, urlencode

import httpx
from httpx._content import encode_request


class TokenDataAuth(httpx.Auth):
    """Pass an access token in a POST body, or as a GET param"""

    ###
    # Since this auth scheme mutates POST bodies to introduce the token, we
    # require access to the body.
    #
    # Ref: https://www.python-httpx.org/advanced/#customizing-authentication
    #
    requires_request_body = True

    def __init__(self, access_token: str):
        self.access_token = access_token

    def auth_flow(self, request: httpx.Request) -> Generator[httpx.Request, httpx.Response, None]:
        if request.method == 'POST':
            self.add_token_to_post_body(request)
        elif request.method == 'GET':
            self.add_token_to_get_params(request)

        yield request

    def add_token_to_post_body(self, request: httpx.Request):
        data = dict(parse_qsl(request.content))
        data['token'] = self.access_token
        headers, request.stream = encode_request(data=data)

        # Ensure our Content-Length header is updated to new value
        request.headers.update(headers)

    def add_token_to_get_params(self, request: httpx.Request):
        params = dict(parse_qsl(request.content))
        params['token'] = self.access_token
        request.url.query = urlencode(params)



"""
Async client for the imagetyperz API

Implementation details from: https://github.com/imagetyperz-api/imagetyperz-api-python3
API Docs: https://www.imagetyperz.com/Forms/api/api.html
Additional docs: https://github.com/imagetyperz-api/API-docs
"""
import asyncio
from datetime import datetime
from typing import Union

import httpx

from .auth import TokenDataAuth
from .constants import reCAPTCHAType
from .exceptions import (raise_on_error, ImageTimedOut, NotDecoded,
                         ImageTyperzError, LimitExceeded, MaximumAttemptsReached)
from .util.logging import getLogger, ClassLoggingProperty

logger = getLogger(__name__)


class ImageTyperzClient:
    session: httpx.AsyncClient
    log = ClassLoggingProperty(logger)

    _access_token: str

    _is_logged_in: bool

    class Endpoints:
        CAPTCHA_CONTENT = 'http://captchatypers.com/Forms/UploadFileAndGetTextNEWToken.ashx'
        CAPTCHA_URL = 'http://captchatypers.com/Forms/FileUploadAndGetTextCaptchaURLToken.ashx'
        RECAPTCHA_SUBMIT = 'http://captchatypers.com/captchaapi/UploadRecaptchaToken.ashx'
        RECAPTCHA_RETRIEVE = 'http://captchatypers.com/captchaapi/GetRecaptchaTextToken.ashx'
        BALANCE = 'http://captchatypers.com/Forms/RequestBalanceToken.ashx'
        BAD_IMAGE = 'http://captchatypers.com/Forms/SetBadImageToken.ashx'
        PROXY_CHECK = 'http://captchatypers.com/captchaAPI/GetReCaptchaTextTokenJSON.ashx'
        GEETEST_SUBMIT = 'http://captchatypers.com/captchaapi/UploadGeeTestToken.ashx'

    def __init__(self, access_token):
        self._access_token = access_token

        self.session = httpx.AsyncClient(auth=TokenDataAuth(self._access_token))

    async def __aenter__(self):
        await self.session.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.session.__aexit__(exc_type, exc_val, exc_tb)

    async def aclose(self):
        """Close the httpx client session"""
        await self.session.aclose()

    def _raise_on_error(self, response: httpx.Response, msg: str, *format_args,
                        no_log_exceptions=(ImageTimedOut, NotDecoded)):
        try:
            raise_on_error(response)
        except ImageTyperzError as e:
            if not isinstance(e, no_log_exceptions):
                self.log.error('%s: %s', msg, *format_args, e)
            raise

    async def retrieve_balance(self) -> float:
        """Retrieve the amount of funds in the authenticated account

        :return:
            Float representing the amount of money available in the account,
            in USD.

        """
        data = {
            'action': 'REQUESTBALANCE',
        }
        res = await self.session.post(self.Endpoints.BALANCE, data=data)
        self._raise_on_error(res, 'Error retrieving account balance')

        return float(res.text)

    async def submit_recaptcha(
        self, *,
        page_url: str,
        site_key: str,
        user_agent: str = None,
        recaptcha_type: Union[int, reCAPTCHAType] = reCAPTCHAType.AUTO,
        recaptcha_action: str = None,
        recaptcha_min_score: float = None,
        data_s: str = None,
    ) -> str:
        """Submit a reCAPTCHA v2 or v3 job

        :param page_url:
            Page URL of website, eg. abc.com

        :param site_key:
            Site key passed to reCAPTCHA JS API, has to be scraped from site

        :param user_agent:
            User-Agent used in solving recaptcha

        :param recaptcha_type:
            Can be one of this 3 values:
                0 - determine automatically (default)
                1 - normal
                2 - invisible
                3 - v3

        :param recaptcha_action:
            Action parameter used in solving v3 recaptcha

        :param recaptcha_min_score:
            Score targeted, check being done against a test recaptcha

        :param data_s:
            Required with some reCAPTCHAs. A one-time token generated with each
            captcha loaded.

        :return:
            ID of the reCAPTCHA job, used to retrieve the response separately,
            afterward.

        """
        data = {
            'action': 'UPLOADCAPTCHA',
            'pageurl': page_url,
            'googlekey': site_key,
            'useragent': user_agent,
            'recaptchatype': recaptcha_type,
            'captchaaction': recaptcha_action,
            'score': recaptcha_min_score,
            'data-s': data_s,
        }
        data = {
            k: v
            for k, v in data.items()
            if v is not None
        }

        self.log.trace('Submitting reCAPTCHA job: %s', data)
        res = await self.session.post(url=self.Endpoints.RECAPTCHA_SUBMIT, data=data)
        self._raise_on_error(res, 'Error submitting reCAPTCHA job')

        job_id = res.text.strip()
        self.log.debug('Successfully submitted reCAPTCHA job: %s', job_id)
        return job_id

    async def retrieve_recaptcha(self, job_id: str) -> str:
        """Retrieve results of a reCAPTCHA job

        :param job_id:
            reCAPTCHA job ID, as returned by submit_captcha()

        :return:
            The g-response of the completed captcha, which may be used to bypass
            the captcha on the website.

        """
        data = {
            'action': 'GETTEXT',
            'captchaid': job_id,
        }

        self.log.trace('Retrieving reCAPTCHA job: %s', data)
        res = await self.session.post(self.Endpoints.RECAPTCHA_RETRIEVE, data=data)
        self._raise_on_error(res, 'Error retrieving reCAPTCHA job %s', job_id)

        g_response = res.text
        self.log.debug('Retrieved reCAPTCHA job %s: %s', job_id, g_response)
        return g_response

    async def complete_recaptcha(
        self, *,
        page_url: str,
        site_key: str,
        user_agent: str = None,
        recaptcha_type: Union[int, reCAPTCHAType] = reCAPTCHAType.AUTO,
        recaptcha_action: str = None,
        recaptcha_min_score: float = None,
        data_s: str = None,
        max_attempts: int = 10,
        poll_interval: float = 2.5,
    ) -> str:
        """Submit reCAPTCHA v2 or v3 job and return its result when available

        This method will await the results of a reCAPTCHA job. If an error of
        IMAGE_TIMED_OUT or LIMIT_EXCEED is returned, the job will be resubmitted
        up to max_attempts-1 times.

        :param page_url:
            Page URL of website, eg. abc.com

        :param site_key:
            Site key passed to reCAPTCHA JS API, has to be scraped from site

        :param user_agent:
            User-Agent used in solving recaptcha

        :param recaptcha_type:
            The version/flavour of the reCAPTCHA. Can be one of:
                0 - determine automatically (default)
                1 - normal
                2 - invisible
                3 - v3

        :param recaptcha_action:
            Action parameter used in solving v3 recaptcha

        :param recaptcha_min_score:
            Score targeted, check being done against a test recaptcha

        :param data_s:
            Required with some reCAPTCHAs. A one-time token generated with each
            captcha loaded.

        :param max_attempts:
            Maximum number of times to submit

        :param poll_interval:
            Time to wait between checking for job results

        :return:
            The g-response of the completed captcha, which may be used to bypass
            the captcha on the website.

        """
        start_at = datetime.utcnow()

        for attempt in range(max_attempts):
            job_id = await self.submit_recaptcha(
                page_url=page_url,
                site_key=site_key,
                user_agent=user_agent,
                recaptcha_type=recaptcha_type,
                recaptcha_action=recaptcha_action,
                recaptcha_min_score=recaptcha_min_score,
                data_s=data_s,
            )

            while True:
                try:
                    g_response = await self.retrieve_recaptcha(job_id)
                except NotDecoded:
                    await asyncio.sleep(poll_interval)
                except (ImageTimedOut, LimitExceeded) as e:
                    attempts_left = max_attempts - attempt - 1
                    if attempts_left > 0:
                        next_step = f'Resubmitting job ({attempts_left} attempts left)'
                    else:
                        next_step = f'Maximum attempts ({max_attempts}) reached. Aborting.'

                    self.log.debug('reCAPTCHA solve job %s expired (%s). %s', job_id, e, next_step)
                    break
                else:
                    end_at = datetime.utcnow()
                    elapsed = end_at - start_at
                    self.log.info('Solved reCAPTCHA job %s in %ss: %s', job_id, elapsed, g_response)
                    return g_response

        else:
            raise MaximumAttemptsReached(
                f'Maximum number of attempts ({max_attempts}) reached while '
                f'trying to solve reCAPTCHA')



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['imagetyperz', 'imagetyperz.util']

package_data = \
{'': ['*']}

modules = \
['LICENSE', 'CHANGELOG']
install_requires = \
['httpx>=0.15.0']

setup_kwargs = {
    'name': 'imagetyperz-async',
    'version': '0.2.0',
    'description': 'Asynchronous client for the ImageTyperz API',
    'long_description': "# imagetyperz-async\n\nAn asynchronous client for the [ImageTyperz CAPTCHA-solving API](http://imagetyperz.com/).\n\n[httpx](https://github.com/encode/httpx) powers the HTTP requests.\n\n**At the moment, only reCAPTCHAs are supported.**\n\n\n# Installation\n\n```bash\npip install imagetyperz-async\n```\n\n\n# Usage\n\n```python\nimport asyncio\n\nfrom imagetyperz import ImageTyperzClient, reCAPTCHAType\nfrom imagetyperz.exceptions import NotDecoded\n\nasync def demo():\n    ###\n    # Context manager will handle the closing of connections in the underlying\n    # httpx AsyncClient at block end.\n    #\n    # Alternatively, `await ita.aclose()` may be manually called to perform\n    # cleanup.\n    #\n    # If no cleanup is performed, a warning may be emitted at Python exit.\n    #\n    async with ImageTyperzClient('6F0848592604C9E14F0EBEA7368493C5') as ita:\n        print(await ita.retrieve_balance())\n        #: 8.8325\n\n        # Submit reCAPTCHA job\n        job_id = await ita.submit_recaptcha(\n            page_url='https://example.com/login',\n            site_key='scraped-site-key',\n            recaptcha_type=reCAPTCHAType.INVISIBLE,\n        )\n        print(job_id)\n        #: 176140709\n\n        # Check for results of the reCAPTCHA job\n        while True:\n            try:\n                g_response = await ita.retrieve_recaptcha(job_id)\n            except NotDecoded:\n                await asyncio.sleep(5)\n                continue\n            else:\n                print(g_response)\n                #: 03AGdBq25hDTCjOq4QywdrY...\n                break\n\n        # Alternatively, use complete_recaptcha to automatically handle the polling\n        # for results — returning with the result when ready.\n        g_response = await ita.complete_recaptcha(\n            page_url='https://example.com/login',\n            site_key='scraped-site-key',\n            recaptcha_type=reCAPTCHAType.INVISIBLE,\n        )\n        print(g_response)\n        #: 03AGdBq25hDTCjOq4QywdrY...\n```\n",
    'author': 'Zach "theY4Kman" Kanzler',
    'author_email': 'they4kman@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/theY4Kman/imagetyperz-async',
    'packages': packages,
    'package_data': package_data,
    'py_modules': modules,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
